"""
Reward function for Microsoft Teams SPA
"""

from typing import Any, Dict, List, Tuple


# === Easy ===
def _validate_respond_to_hermione(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function that checks if the user has sent the message
    "Sure thing - it looks good!" to Hermione.

    Args:
        state: The current SPA state as a dictionary

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if message sent, 0.0 otherwise
    """
    user_id = "you"
    target_conversation_id = "hermione-granger"
    expected_content = "Sure thing - it looks good!"

    messages: List[Dict[str, Any]] = final_state.get("messages", [])

    # Find Hermione's conversation
    hermione_convo = None
    for convo in messages:
        if convo.get("conversationId") == target_conversation_id:
            hermione_convo = convo
            break

    if hermione_convo is None:
        return 0.0, "Hermione's conversation not found."

    # Check if the expected message exists in the conversation
    for msg in hermione_convo.get("messages", []):
        if msg.get("senderId") == user_id and msg.get("content") == expected_content:
            return 1.0, "Success! The expected message was sent to Hermione."

    return 0.0, f"Message not sent. Expected: '{expected_content}'"


def _validate_message_tim_white_twice(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function that checks if the user has sent two specific messages
    to Tim White.

    Args:
        initial_state: The initial SPA state as a dictionary
        final_state: The final SPA state as a dictionary

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if both messages sent, 0.0 otherwise
    """
    user_id = "you"
    target_conversation_id = "tim-white"
    expected_messages = ["Hey Tim, smoke test completed.", "Let me know if you need a hand with anything."]

    messages: List[Dict[str, Any]] = final_state.get("messages", [])

    # Find Tim White's conversation
    tim_convo = None
    for convo in messages:
        if convo.get("conversationId") == target_conversation_id:
            tim_convo = convo
            break

    if tim_convo is None:
        return 0.0, "Tim White's conversation not found."

    # Check if both expected messages exist in the conversation
    sent_messages = [msg.get("content") for msg in tim_convo.get("messages", []) if msg.get("senderId") == user_id]
    missing = [msg for msg in expected_messages if msg not in sent_messages]

    if not missing:
        return 1.0, "Success! Both messages were sent to Tim White."
    else:
        return 0.0, f"Message(s) not sent: {missing}"


def _validate_harry_hagrid_harry_hagrid(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function that checks if the user completed the sequence:
      1. Messaged Harry: "Hey Harry, can we reschedule to 2? I have a meeting with Hagrid."
      2. Messaged Hagrid: "Hey Hagrid, should be all good - I'll let you know asap."
      3. Messaged Harry: "Good to hear - see you at 2!"
      4. Messaged Hagrid: "Harry said yes, I'll see you at 12 then!"

    Args:
        initial_state: The starting SPA state
        final_state: The final SPA state after user actions

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if all messages are correctly sent, 0.0 otherwise
    """
    user_id = "you"
    harry_convo_id = "harry-potter"
    hagrid_convo_id = "hagrid-cao"

    expected_messages = {
        harry_convo_id: ["Hey Harry, can we reschedule to 2? I have a meeting with Hagrid.", "Good to hear - see you at 2!"],
        hagrid_convo_id: [
            "Hey Hagrid, should be all good - I'll let you know asap.",
            "Harry said yes, I'll see you at 12 then!",
        ],
    }

    messages: List[Dict[str, Any]] = final_state.get("messages", [])
    found: dict[str, list[str]] = {harry_convo_id: [], hagrid_convo_id: []}

    for convo in messages:
        convo_id = convo.get("conversationId")
        if convo_id in expected_messages:
            for msg in convo.get("messages", []):
                if msg.get("senderId") == user_id:
                    content = msg.get("content", "")
                    if content in expected_messages[convo_id]:
                        found[convo_id].append(content)

    harry_msgs = set(found[harry_convo_id])
    hagrid_msgs = set(found[hagrid_convo_id])

    harry_ok = harry_msgs == set(expected_messages[harry_convo_id])
    hagrid_ok = hagrid_msgs == set(expected_messages[hagrid_convo_id])

    if harry_ok and hagrid_ok:
        return 1.0, "Success! All messages were correctly sent to Harry and Hagrid in both chats."

    missing: list[str] = []
    for convo_id, expected in expected_messages.items():
        convo_name = "Harry" if convo_id == harry_convo_id else "Hagrid"
        missing_msgs = [m for m in expected if m not in found[convo_id]]
        if missing_msgs:
            missing.append(f"{convo_name}: {missing_msgs}")

    return 0.0, f"Some messages are missing or incorrect: {', '.join(missing)}"


def _validate_open_search_page_for_hagrid(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function that checks if the user opened the search page for Hagrid.

    Args:
        initial_state: The SPA state before user action
        final_state: The SPA state after user action

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if the search page shows results for Hagrid, 0.0 otherwise
    """
    expected_search_term = "hagrid"
    expected_page = "SEARCH"
    expected_search_page_type = "PEOPLE"

    # Check if the search page is open
    page = final_state.get("page")
    search_page = final_state.get("searchPage")
    search_term = final_state.get("searchTerm", "").lower()

    if page != expected_page:
        return 0.0, f"Wrong page open: {page}. Expected HOME → SEARCH."

    if search_page != expected_search_page_type:
        return 0.0, f"Search page type is '{search_page}', expected '{expected_search_page_type}'."

    if search_term != expected_search_term:
        return 0.0, f"Search term is '{search_term}', expected '{expected_search_term}'."

    return 1.0, "Success! Hagrid's search page is open and ready."


def _validate_prepare_to_make_a_new_chat(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function that checks if the user has selected the new chat icon
    and entered the new chat screen.

    Args:
        initial_state: The SPA state before the action
        final_state: The SPA state after the action

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if the new chat screen is shown, 0.0 otherwise
    """
    # Check if the SPA has transitioned to creating a new chat
    creating_chat = final_state.get("creatingChat", False)

    if creating_chat:
        return 1.0, "Success! The new chat screen is now shown."

    return 0.0, "New chat screen not shown. Make sure the new chat icon was clicked."


def _validate_filter_chats_for_harry(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function that checks if the sidebar chats are filtered
    for the name 'harry', meaning only Harry Potter's chat is shown.

    Args:
        state: The current SPA state as a dictionary.

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if only Harry Potter's chat is displayed after filtering,
                 0.0 otherwise.
    """
    target_filter = "harry"

    # Extract filtering info
    filter_string: str = final_state.get("filterString", "")
    filtering: bool = final_state.get("filtering", False)
    # messages: List[Dict[str, Any]] = final_state.get("messages", [])

    # If filtering not active or filter string is wrong
    if not filtering:
        return 0.0, "Filtering is not active."
    if filter_string.lower() != target_filter:
        return 0.0, f"Expected filter string '{target_filter}', got '{filter_string}'."

    return 1.0, "Success! Filters for Harry potter's chat."


def _validate_navigate_to_harrys_files_page(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function that checks if the user has navigated to Harry's files page.

    Args:
        initial_state: The initial SPA state as a dictionary
        final_state: The final SPA state as a dictionary

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if files page is open, 0.0 otherwise
    """
    expected_page = "FILES"
    current_page = final_state.get("mainChatPage", "")

    if current_page == expected_page:
        return 1.0, "Success! The files page is open."

    return 0.0, f"Files page not open. Current page: '{current_page}'"


def _validate_navigate_to_harrys_photos_page(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function that checks if the user has navigated to Harry's photos page.

    Args:
        initial_state: The initial SPA state as a dictionary
        final_state: The final SPA state as a dictionary

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if files page is open, 0.0 otherwise
    """
    expected_page = "PHOTOS"
    current_page = final_state.get("mainChatPage", "")

    if current_page == expected_page:
        return 1.0, "Success! The photos page is open."

    return 0.0, f"Photos page not open. Current page: '{current_page}'"


# === Medium ===
def _validate_message_hermione_and_then_tim(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function that checks if the user has:
    1. Sent Hermione the message "All done. Let me know if you get it!"
    2. Sent Tim the message "Smoke tested! Should be with you shortly."

    Args:
        initial_state: The initial SPA state
        final_state: The final SPA state after user interaction

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if both messages were sent correctly, 0.5 if only one was sent, 0.0 otherwise
    """
    user_id = "you"

    # Define targets
    hermione_conversation_id = "hermione-granger"
    tim_conversation_id = "tim-white"

    expected_hermione_msg = "All done. Let me know if you get it!"
    expected_tim_msg = "Smoke tested! Should be with you shortly."

    messages: List[Dict[str, Any]] = final_state.get("messages", [])

    # Helper to find a conversation by ID
    def find_convo(convo_id: str) -> Dict[str, Any]:
        for convo in messages:
            if convo.get("conversationId") == convo_id:
                return convo
        return {}

    # Check Hermione convo
    hermione_convo = find_convo(hermione_conversation_id)
    hermione_msg_sent = (
        any(
            msg.get("senderId") == user_id and msg.get("content") == expected_hermione_msg
            for msg in hermione_convo.get("messages", [])
        )
        if hermione_convo
        else False
    )

    # Check Tim convo
    tim_convo = find_convo(tim_conversation_id)
    tim_msg_sent = (
        any(msg.get("senderId") == user_id and msg.get("content") == expected_tim_msg for msg in tim_convo.get("messages", []))
        if tim_convo
        else False
    )

    # Reward logic
    if hermione_msg_sent and tim_msg_sent:
        return 1.0, "Success! Messages sent to both Hermione and Tim."
    elif hermione_msg_sent or tim_msg_sent:
        missing = "Tim" if hermione_msg_sent else "Hermione"
        return 0.5, f"Partially complete. Missing message to {missing}."
    else:
        return 0.0, "Messages not sent to either Hermione or Tim."


def _validate_make_a_group_chat_with_hagrid_and_ron(
    initial_state: Dict[str, Any], final_state: Dict[str, Any]
) -> Tuple[float, str]:  # noqa: E501
    """
    Reward function that checks if the user has created a group chat
    including Hagrid and Ron, and sent the message "Hey guys!".

    Args:
        initial_state: The initial SPA state as a dictionary
        final_state: The final SPA state as a dictionary

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if the correct group chat was created and message sent, 0.0 otherwise
    """
    user_id = "you"
    expected_participants = {"hagrid-cao", "ron-weasley"}
    expected_message = "Hey guys!"

    messages: List[Dict[str, Any]] = final_state.get("messages", [])
    if not messages:
        return 0.0, "No messages found in final state."

    # Look for a new conversation containing both Hagrid and Ron
    for convo in messages:
        user_ids = set(convo.get("userIds", []))
        if expected_participants.issubset(user_ids):
            # Check that the user sent "Hey guys!" in this conversation
            for msg in convo.get("messages", []):
                if msg.get("senderId") == user_id and msg.get("content") == expected_message:
                    return 1.0, "Success! Created group chat with Hagrid and Ron and sent 'Hey guys!'."
            return 0.0, "Group chat found, but expected message not sent."

    return 0.0, "No group chat found with both Hagrid and Ron."


def _validate_create_a_chat_with_fred_weasely(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:  # noqa: E501
    """
    Reward function that checks if a new chat was created with Fred Weasley
    and the message "Hey Fred, how are you!" was sent.

    Args:
        state: The current SPA state as a dictionary.

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if chat created and message sent, 0.0 otherwise.
    """
    user_id = "you"
    target_conversation_id = "fred-weasley"
    expected_content = "Hey Fred, how are you!"

    messages: List[Dict[str, Any]] = final_state.get("messages", [])

    # Find Fred's conversation
    fred_convo = None
    for convo in messages:
        if convo.get("conversationId") == target_conversation_id:
            fred_convo = convo
            break

    if fred_convo is None:
        return 0.0, "Conversation with Fred Weasley not found. Make sure you started a new chat."

    # Check if the expected message exists in the conversation
    for msg in fred_convo.get("messages", []):
        if msg.get("senderId") == user_id and msg.get("content") == expected_content:
            return 1.0, "Success! Chat with Fred Weasley created and message sent."

    return 0.0, f"Message not found in Fred's chat. Expected: '{expected_content}'"


def _validate_message_tim_and_select_hermiones_photos(
    initial_state: Dict[str, Any], final_state: Dict[str, Any]
) -> Tuple[float, str]:  # noqa: E501
    """
    Reward function that validates whether:
      1. The user sent "Hey Tim, it's me" to Tim White.
      2. The user navigated to Hermione's chat and selected the Photos tab.

    Args:
        initial_state: The SPA's starting state as a dictionary.
        final_state: The SPA's ending state as a dictionary.

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if both actions are complete, 0.5 if only message is sent,
                 0.0 otherwise.
    """

    user_id = "you"
    tim_conversation_id = "tim-white"
    hermione_conversation_id = "hermione-granger"
    expected_message = "Hey Tim, it's me"

    messages: List[Dict[str, Any]] = final_state.get("messages", [])

    # --- Step 1: Check message to Tim ---
    tim_convo = None
    for convo in messages:
        if convo.get("conversationId") == tim_conversation_id:
            tim_convo = convo
            break

    message_sent = False
    if tim_convo:
        for msg in tim_convo.get("messages", []):
            if msg.get("senderId") == user_id and msg.get("content") == expected_message:
                message_sent = True
                break

    # --- Step 2: Check Hermione's chat and Photos tab ---
    active_convo = final_state.get("msgId")
    active_tab = final_state.get("mainChatPage")

    on_hermione_photos = active_convo == hermione_conversation_id and active_tab == "PHOTOS"

    # --- Combine results ---
    if message_sent and on_hermione_photos:
        return 1.0, "Success! Message sent to Tim and Hermione's Photos tab is open."
    elif message_sent:
        return 0.5, "Partial success. Message sent to Tim, but not on Hermione's Photos tab."
    else:
        return 0.0, "Message to Tim not sent yet, and Hermione's Photos tab not open."


def _validate_search_for_messages_containing_hey(
    initial_state: Dict[str, Any], final_state: Dict[str, Any]
) -> Tuple[float, str]:  # noqa: E501
    """
    Reward function that checks if a search for 'hey' was performed and
    if the result containing "Hey Hugo — are we still on for Friday at 12pm?" is shown.

    Args:
        initial_state: The initial SPA state as a dictionary
        final_state: The final SPA state as a dictionary

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if search result is present, 0.0 otherwise
    """
    expected_message_content = "Hey Hugo — are we still on for Friday at 12pm?"

    # Check if the user navigated to the MESSAGES tab in the search page
    if final_state.get("page") != "SEARCH" or final_state.get("searchPage") != "MESSAGES":
        return 0.0, "User did not navigate to the messages tab after searching."

    # Look through all messages in the final state
    messages: List[Dict[str, Any]] = final_state.get("messages", [])
    for convo in messages:
        for msg in convo.get("messages", []):
            if expected_message_content in msg.get("content", ""):
                return 1.0, f"Success! Found expected message: '{expected_message_content}'"

    return 0.0, f"Message not found. Expected: '{expected_message_content}'"


def _validate_message_tim_and_alex_langford(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function that checks if the user sent the message
    "Smoke test finished, looks good!" to Tim and also started a
    new conversation with Alex Langford, sending "Good to see you Alex!".

    Args:
        initial_state: The initial SPA state as a dictionary
        final_state: The final SPA state as a dictionary

    Returns:
        Tuple[float, str]:
            reward = 1.0 if both conditions met,
                     0.5 if only one of them is met,
                     0.0 otherwise.
            message = explanation of result.
    """
    user_id = "you"
    messages: List[Dict[str, Any]] = final_state.get("messages", [])

    # --- Targets ---
    tim_convo_id = "tim-white"
    tim_expected_content = "Smoke test finished, looks good!"

    alex_convo_id = "alex-langford"
    alex_expected_content = "Good to see you Alex!"

    tim_message_sent = False
    alex_message_sent = False

    # Check Tim’s conversation
    for convo in messages:
        if convo.get("conversationId") == tim_convo_id:
            for msg in convo.get("messages", []):
                if msg.get("senderId") == user_id and msg.get("content") == tim_expected_content:
                    tim_message_sent = True
                    break

    # Check Alex’s conversation (it must exist and have the message)
    for convo in messages:
        if convo.get("conversationId") == alex_convo_id:
            for msg in convo.get("messages", []):
                if msg.get("senderId") == user_id and msg.get("content") == alex_expected_content:
                    alex_message_sent = True
                    break

    # --- Scoring logic ---
    if tim_message_sent and alex_message_sent:
        return 1.0, "Success! Messages sent to both Tim and Alex."
    elif tim_message_sent:
        return 0.5, "Partial success: message sent to Tim, but not to Alex."
    elif alex_message_sent:
        return 0.5, "Partial success: message sent to Alex, but not to Tim."
    else:
        return 0.0, "No valid messages sent to Tim or Alex."


# === Hard ===
def _validate_large_group_chat(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function that checks if the user has created a group chat with
    Harry Potter, Hermione Granger, Tim White, and Tim White 2,
    and sent the message 'What a group chat!'.

    Args:
        initial_state: The initial SPA state as a dictionary
        final_state: The final SPA state as a dictionary

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if the group chat and message exist, 0.0 otherwise
    """
    user_id = "you"
    expected_users = {"harry-potter", "hermione-granger", "tim-white", "tim-white-2"}
    expected_message = "What a group chat!"

    messages: List[Dict[str, Any]] = final_state.get("messages", [])

    # Find a conversation that has exactly the expected users
    group_convo = None
    for convo in messages:
        convo_users = set(convo.get("userIds", []))
        if convo_users == expected_users:
            group_convo = convo
            break

    if group_convo is None:
        return 0.0, "Group chat with the required users not found."

    # Check if the expected message was sent by the user in that chat
    for msg in group_convo.get("messages", []):
        if msg.get("senderId") == user_id and msg.get("content") == expected_message:
            return 1.0, "Success! Group chat created and message sent."

    return 0.0, f"Message not found in the group chat. Expected: '{expected_message}'"


def _validate_message_hermione_and_search_for_it(
    initial_state: Dict[str, Any], final_state: Dict[str, Any]
) -> Tuple[float, str]:  # noqa: E501
    """
    Reward function that verifies:
    1. The user sent the message "All good Hermione, it's finished." to Hermione.
    2. The user performed a search for 'Hermi' and opened the 'messages' tab.

    Args:
        initial_state: The initial SPA state as a dictionary.
        final_state: The final SPA state as a dictionary.

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if both conditions are met, partial reward if only one is met.
    """
    user_id = "you"
    target_conversation_id = "hermione-granger"
    expected_content = "All good Hermione, it's finished."
    expected_search_term = "Hermi"
    expected_search_page = "MESSAGES"

    messages: List[Dict[str, Any]] = final_state.get("messages", [])

    # Step 1: Verify Hermione's conversation and message
    hermione_convo = None
    for convo in messages:
        if convo.get("conversationId") == target_conversation_id:
            hermione_convo = convo
            break

    if hermione_convo is None:
        return 0.0, "Hermione's conversation not found."

    sent_message = any(
        msg.get("senderId") == user_id and msg.get("content") == expected_content for msg in hermione_convo.get("messages", [])
    )

    # Step 2: Verify search term and tab
    search_term = final_state.get("searchTerm", "")
    search_page = final_state.get("searchPage", "")
    on_search_page = final_state.get("page") == "SEARCH"

    correct_search = on_search_page and search_term == expected_search_term and search_page == expected_search_page

    # Evaluate total reward
    if sent_message and correct_search:
        return 1.0, "Success! You messaged Hermione and searched for it correctly."
    elif sent_message:
        return 0.5, "Message sent, but search not completed correctly."
    elif correct_search:
        return 0.5, "Search was correct, but Hermione’s message wasn’t sent."
    else:
        return 0.0, "Neither the message nor the search action were completed correctly."


def _validate_message_the_third_user_in_the_search(
    initial_state: Dict[str, Any], final_state: Dict[str, Any]
) -> Tuple[float, str]:  # noqa: E501
    """
    Reward function that checks if the user has messaged the third user in the
    search list ("Anne Peters") with the message "Hi Anne, it's me."

    Args:
        initial_state: The starting SPA state as a dictionary
        final_state: The resulting SPA state as a dictionary

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if message sent correctly, 0.0 otherwise
    """
    user_id = "you"
    target_conversation_id = "anne-peters"
    expected_content = "Hi Anne, it's me."

    messages: List[Dict[str, Any]] = final_state.get("messages", [])

    # Locate Anne's conversation
    anne_convo = None
    for convo in messages:
        if convo.get("conversationId") == target_conversation_id:
            anne_convo = convo
            break

    if anne_convo is None:
        return 0.0, "Anne's conversation not found — looks like you didn't message her."

    # Check if the expected message exists
    for msg in anne_convo.get("messages", []):
        if msg.get("senderId") == user_id and msg.get("content") == expected_content:
            return 1.0, "Nice! You successfully messaged Anne Peters with the correct text."

    return 0.0, f"Message not sent or incorrect. Expected: '{expected_content}'"


# _validate_harry_hagrid_harry_hagrid
def _validate_message_and_search_for_2(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function for verifying that:
    1. The user sent two specific messages to Harry Potter.
    2. The user searched for the number '2' in messages.
    3. The search results display five messages (four from user, one from Harry).

    Args:
        initial_state: The starting state of the SPA
        final_state: The ending state after the user's actions

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if all criteria are met, 0.0 otherwise
    """
    user_id = "you"
    target_conversation_id = "harry-potter"

    required_messages = ["What do you mean you said we'd meet at two?!", "Wait let me check my message history"]

    # 1. Check messages to Harry
    messages: List[Dict[str, Any]] = final_state.get("messages", [])
    harry_convo = next((convo for convo in messages if convo.get("conversationId") == target_conversation_id), None)

    if harry_convo is None:
        return 0.0, "Harry's conversation not found."

    convo_messages = harry_convo.get("messages", [])
    sent_texts = [m.get("content") for m in convo_messages if m.get("senderId") == user_id]

    if not all(req in sent_texts for req in required_messages):
        return 0.0, "Missing one or more required messages to Harry."

    # 2. Check search parameters
    search_term = final_state.get("searchTerm", "")
    search_page = final_state.get("searchPage", "")

    if search_term != "2" or search_page != "MESSAGES":
        return (
            0.0,
            f"Search not completed correctly. Expected term='2' and page='MESSAGES', got term='{search_term}', page='{search_page}'.",  # noqa: E501
        )  # noqa: E501

    # 3. Check number of messages in search results (five total, four from user, one from Harry)
    harry_msgs = [
        msg for convo in messages if convo.get("conversationId") == target_conversation_id for msg in convo.get("messages", [])
    ]

    num_from_user = sum(1 for m in harry_msgs if m.get("senderId") == user_id)
    num_from_harry = sum(1 for m in harry_msgs if m.get("senderId") == "harry-potter")

    if num_from_user == 5 and num_from_harry == 1:  # three to harry, one to hermione and tim.
        return 1.0, "Success! Messages sent, search performed, and correct results displayed."
    else:
        return 0.0, f"Incorrect number of messages displayed. Found {num_from_user} from you and {num_from_harry} from Harry."


def _validate_the_a_team(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function for the "The A team" task in Microsoft Teams.
    Checks if a group chat with Alex Langford, Alex Langford II, and Anne Peters
    exists and contains the message "The A team" sent by the user.

    Args:
        initial_state: The initial SPA state (not used here)
        final_state: The final SPA state

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if group chat exists and message sent, 0.0 otherwise
    """
    user_id = "you"
    target_users = {"alex-langford", "alex-langford-2", "anne-peters"}
    expected_message = "The A team"

    messages: List[Dict[str, Any]] = final_state.get("messages", [])

    # Find the group chat containing all three target users
    target_chat = None
    for convo in messages:
        convo_users = set(["alex-langford", "alex-langford-2", "anne-peters"])
        if target_users.issubset(convo_users):
            target_chat = convo
            break

    if target_chat is None:
        return 0.0, "Group chat with Alex Langford, Alex Langford II, and Anne Peters not found."

    # Check if the expected message was sent by the user
    for msg in target_chat.get("messages", []):
        if msg.get("senderId") == user_id and msg.get("content") == expected_message:
            return 1.0, "Success! The group chat was created and message sent."

    return 0.0, f"Group chat exists but message '{expected_message}' was not sent."


def _validate_birthday_for_george(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function that checks if the user has:
      1. Created a group chat with Fred Weasley, Hermione Granger, and Harry Potter
      2. Sent the correct message in that group chat
      3. Sent a follow-up message to George Weasley
      4. Typed "Ron Weasley" into the search bar (without pressing enter)
      5. Stayed on the HOME page

    Args:
        state: The current SPA state as a dictionary

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if all criteria are met, 0.0 otherwise
    """

    user_id = "you"
    messages: List[Dict[str, Any]] = final_state.get("messages", [])
    page: str = final_state.get("page", "")
    search_term: str = final_state.get("searchTerm", "").strip().lower()

    # --- 1. Check for group chat ---
    expected_group_users = {"fred-weasley", "harry-potter", "hermione-granger"}
    group_message_text = "Hey guys, just confirming we're throwing a surprise party for George tomorrow at 3. See you then!"

    group_chat_found = False
    group_message_ok = False
    george_message_ok = False

    for convo in messages:
        user_ids = set(convo.get("userIds", []))
        convo_messages = convo.get("messages", [])

        # Check group chat
        if user_ids == expected_group_users:
            group_chat_found = True
            for msg in convo_messages:
                if msg.get("senderId") == user_id and msg.get("content") == group_message_text:
                    group_message_ok = True
                    break

        # Check message to George
        if user_ids == {"george-weasley"}:
            for msg in convo_messages:
                if (
                    msg.get("senderId") == user_id
                    and msg.get("content") == "Hey George, is it your day off tomorrow? I can't remember if you're out."
                ):
                    george_message_ok = True
                    break

    # --- 2. Check all criteria ---
    if not group_chat_found:
        return 0.0, "Group chat with Fred, Harry, and Hermione not found."

    if not group_message_ok:
        return 0.0, "Expected message to the group chat not found."

    if not george_message_ok:
        return 0.0, "Expected message to George Weasley not found."

    if search_term != "ron weasley":
        return 0.0, f"Search term incorrect. Expected 'Ron Weasley', found '{final_state.get('searchTerm', '')}'."

    if page != "HOME":
        return 0.0, f"User is not on HOME page (found '{page}')."

    return 1.0, "Success! All expected actions were completed correctly."


def _validate_group_chat_and_hermione(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Reward function for validating that:
      1. The user created a group chat with Harry and Hagrid and sent the message:
         "I'll see you both at 2. Looking forward to it!"
      2. Then messaged Hermione saying:
         "Hey Hermione, me, Harry and Hagrid are meeting up if you want to join us!"

    Args:
        state: The current SPA state as a dictionary

    Returns:
        Tuple of (reward, message)
        reward = 1.0 if both conditions met, 0.5 if only group chat message sent,
                 0.0 otherwise
    """
    user_id = "you"
    messages: List[Dict[str, Any]] = final_state.get("messages", [])

    # --- Condition 1: Group chat with Harry and Hagrid ---
    expected_group_users = {"hagrid-cao", "harry-potter"}
    expected_group_message = "I'll see you both at 2. Looking forward to it!"
    group_chat_found = False
    group_message_sent = False

    for convo in messages:
        convo_users = set(convo.get("userIds", []))
        if convo_users == expected_group_users:
            group_chat_found = True
            for msg in convo.get("messages", []):
                if msg.get("senderId") == user_id and msg.get("content") == expected_group_message:
                    group_message_sent = True
                    break
            break

    if not group_chat_found:
        return 0.0, "Group chat with Harry and Hagrid not found."

    if not group_message_sent:
        return 0.0, f"Group chat found, but expected message '{expected_group_message}' not sent."

    # --- Condition 2: Message Hermione ---
    hermione_convo = next((c for c in messages if c.get("conversationId") == "hermione-granger"), None)
    expected_hermione_message = "Hey Hermione, me, Harry and Hagrid are meeting up if you want to join us!"
    hermione_message_sent = False

    if hermione_convo is not None:
        for msg in hermione_convo.get("messages", []):
            if msg.get("senderId") == user_id and msg.get("content") == expected_hermione_message:
                hermione_message_sent = True
                break
    else:
        return 0.5, "Hermione's conversation not found, but group chat message was sent."

    if group_message_sent and hermione_message_sent:
        return 1.0, "Success! Both group chat message and Hermione message were sent."
    elif group_message_sent:
        return 0.5, "Partial success — group chat message sent, but Hermione message missing."
    else:
        return 0.0, "Messages missing or incorrect."

    # Registry of all microsoft teams reward functions


REWARD_FUNCTIONS_MICROSOFT_TEAMS = {
    "_validate_respond_to_hermione": _validate_respond_to_hermione,
    "_validate_message_tim_white_twice": _validate_message_tim_white_twice,
    "_validate_prepare_to_make_a_new_chat": _validate_prepare_to_make_a_new_chat,
    "_validate_open_search_page_for_hagrid": _validate_open_search_page_for_hagrid,
    "_validate_filter_chats_for_harry": _validate_filter_chats_for_harry,
    "_validate_navigate_to_harrys_files_page": _validate_navigate_to_harrys_files_page,
    "_validate_navigate_to_harrys_photos_page": _validate_navigate_to_harrys_photos_page,
    "_validate_message_hermione_and_then_tim": _validate_message_hermione_and_then_tim,
    "_validate_make_a_group_chat_with_hagrid_and_ron": _validate_make_a_group_chat_with_hagrid_and_ron,
    "_validate_create_a_chat_with_fred_weasely": _validate_create_a_chat_with_fred_weasely,
    "_validate_message_tim_and_select_hermiones_photos": _validate_message_tim_and_select_hermiones_photos,
    "_validate_search_for_messages_containing_hey": _validate_search_for_messages_containing_hey,
    "_validate_message_tim_and_alex_langford": _validate_message_tim_and_alex_langford,
    "_validate_large_group_chat": _validate_large_group_chat,
    "_validate_message_hermione_and_search_for_it": _validate_message_hermione_and_search_for_it,
    "_validate_message_the_third_user_in_the_search": _validate_message_the_third_user_in_the_search,
    "_validate_harry_hagrid_harry_hagrid": _validate_harry_hagrid_harry_hagrid,
    "_validate_message_and_search_for_2": _validate_message_and_search_for_2,
    "_validate_the_a_team": _validate_the_a_team,
    "_validate_birthday_for_george": _validate_birthday_for_george,
    "_validate_group_chat_and_hermione": _validate_group_chat_and_hermione,
}



from abc import ABC, abstractmethod
from typing import Any, Dict


class Backend(ABC):
    @abstractmethod
    def query(self, query: dict[str, Any]) -> Any:
        """Execute a MongoDB query.
        
        Args:
            query: A dict with 'collection' and 'filter' keys, e.g.
                   {"collection": "users", "filter": {"_id": "0"}}
        
        Returns:
            List of matching documents (or single document if applicable)
        """
        pass

class BackendDictAdapter(Backend):
    """Adapter to wrap backend state dict as a Backend object."""

    def __init__(self, backend_state: Dict[str, Any]):
        self.backend_state = backend_state

    def query(self, query: dict[str, Any]) -> Any:
        collection = query.get("collection")
        filter_dict = query.get("filter", {})

        if collection not in self.backend_state:
            return []

        collection_data = self.backend_state[collection]
        if not isinstance(collection_data, list):
            return []

        results = []
        for item in collection_data:
            if not isinstance(item, dict):
                continue
            match = True
            for key, value in filter_dict.items():
                if isinstance(value, dict) and "$in" in value:
                    if item.get(key) not in value.get("$in", []):
                        match = False
                        break
                else:
                    if item.get(key) != value:
                        match = False
                        break
            if match:
                results.append(item)

        return results


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dojo_sdk_core.dojos.dojos_loader import collect_tasks, load_dojos
    from dojo_sdk_core.dojos.model import Dojo

__all__ = ["Dojo", "collect_tasks", "load_dojos"]


def __getattr__(name: str):
    if name in {"collect_tasks", "load_dojos"}:
        from dojo_sdk_core.dojos import dojos_loader

        return getattr(dojos_loader, name)
    if name == "Dojo":
        from dojo_sdk_core.dojos.model import Dojo

        return Dojo
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""
Reward functions for Outlook app tasks.
"""

import logging
from typing import Any, Dict, Tuple

logger = logging.getLogger(__name__)


def _validate_go_to_the_drafts_page(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully navigated to the Drafts page.

    This function checks:
    1. The currentPage is "drafts"

    Args:
        initial_state: The initial state before navigation
        final_state: The final state after navigation

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check: currentPage should be "drafts"
        current_page = final_state.get("currentPage")
        if current_page != "drafts":
            return (
                0.0,
                f"Expected currentPage to be 'drafts', got '{current_page}'",
            )

        return 1.0, "Successfully navigated to Drafts page"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_go_to_the_sent_items_page(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully navigated to the Sent Items page.

    This function checks:
    1. The currentPage is "sent"

    Args:
        initial_state: The initial state before navigation
        final_state: The final state after navigation

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check: currentPage should be "sent"
        current_page = final_state.get("currentPage")
        if current_page != "sent":
            return (
                0.0,
                f"Expected currentPage to be 'sent', got '{current_page}'",
            )

        return 1.0, "Successfully navigated to Sent Items page"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_open_an_email_from_the_list(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully opened an email from the list.

    This function checks:
    1. The selectedEmailId is "3" (LinkedIn email)
    2. Email with id "3" has read status set to true

    Args:
        initial_state: The initial state before opening
        final_state: The final state after opening

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check 1: selectedEmailId should be "3"
        selected_email_id = final_state.get("selectedEmailId")
        if selected_email_id != "3":
            return (
                0.0,
                f"Expected selectedEmailId to be '3', got '{selected_email_id}'",
            )

        # Check 2: Email with id "3" should have read = true
        emails = final_state.get("emails", [])
        email_3 = next((e for e in emails if e.get("id") == "3"), None)

        if email_3 is None:
            return 0.0, "Email with id '3' not found in emails array"

        if not email_3.get("read"):
            return (
                0.0,
                f"Expected email '3' to have read=true, got read={email_3.get('read')}",
            )

        return 1.0, "Successfully opened email from the list"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_delete_an_email_from_the_list(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully deleted an email from the list.

    This function checks:
    1. Email with id "3" (LinkedIn) is removed from the emails array

    Args:
        initial_state: The initial state before deletion
        final_state: The final state after deletion

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check: Email with id "3" should not exist in emails array
        emails = final_state.get("emails", [])
        email_3 = next((e for e in emails if e.get("id") == "3"), None)

        if email_3 is not None:
            return 0.0, "Email with id '3' (LinkedIn) still exists, should be deleted"

        return 1.0, "Successfully deleted email from the list"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_flag_an_email(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully flagged an email.

    This function checks:
    1. Email with id "3" (LinkedIn) has flagged status set to true

    Args:
        initial_state: The initial state before flagging
        final_state: The final state after flagging

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check: Email with id "3" should have flagged = true
        emails = final_state.get("emails", [])
        email_3 = next((e for e in emails if e.get("id") == "3"), None)

        if email_3 is None:
            return 0.0, "Email with id '3' not found in emails array"

        if not email_3.get("flagged"):
            return (
                0.0,
                f"Expected email '3' to have flagged=true, got flagged={email_3.get('flagged')}",
            )

        return 1.0, "Successfully flagged the email"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_pin_an_email(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully pinned an email.

    This function checks:
    1. Email with id "6" (Google Calendar) has pinned status set to true

    Args:
        initial_state: The initial state before pinning
        final_state: The final state after pinning

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check: Email with id "6" should have pinned = true
        emails = final_state.get("emails", [])
        email_6 = next((e for e in emails if e.get("id") == "6"), None)

        if email_6 is None:
            return 0.0, "Email with id '6' not found in emails array"

        if not email_6.get("pinned"):
            return (
                0.0,
                f"Expected email '6' to have pinned=true, got pinned={email_6.get('pinned')}",
            )

        return 1.0, "Successfully pinned the email"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_toggle_read_on_an_email(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully toggled read status on an email.

    This function checks:
    1. Email with id "3" (LinkedIn) has read status set to true

    Args:
        initial_state: The initial state before toggling
        final_state: The final state after toggling

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check: Email with id "3" should have read = true
        emails = final_state.get("emails", [])
        email_3 = next((e for e in emails if e.get("id") == "3"), None)

        if email_3 is None:
            return 0.0, "Email with id '3' not found in emails array"

        if not email_3.get("read"):
            return (
                0.0,
                f"Expected email '3' to have read=true, got read={email_3.get('read')}",
            )

        return 1.0, "Successfully toggled read status on the email"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_toggle_unread_on_an_email(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully toggled unread status on an email.

    This function checks:
    1. Email with id "2" (Netflix) has read status set to false

    Args:
        initial_state: The initial state before toggling
        final_state: The final state after toggling

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check: Email with id "2" should have read = false
        emails = final_state.get("emails", [])
        email_2 = next((e for e in emails if e.get("id") == "2"), None)

        if email_2 is None:
            return 0.0, "Email with id '2' not found in emails array"

        if email_2.get("read"):
            return (
                0.0,
                f"Expected email '2' to have read=false, got read={email_2.get('read')}",
            )

        return 1.0, "Successfully toggled unread status on the email"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_compose_a_new_draft(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully composed a new draft.

    This function checks:
    1. A new draft email exists (id starts with "draft-")
    2. The currentPage is "drafts"
    3. The selectedEmailId is set to the new draft

    Args:
        initial_state: The initial state before composing
        final_state: The final state after composing

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check 1: currentPage should be "drafts"
        current_page = final_state.get("currentPage")
        if current_page != "drafts":
            return (
                0.0,
                f"Expected currentPage to be 'drafts', got '{current_page}'",
            )

        # Check 2: Find a new draft email
        emails = final_state.get("emails", [])
        new_draft = next((e for e in emails if e.get("id", "").startswith("draft-") and e.get("folder") == "drafts"), None)

        if new_draft is None:
            return 0.0, "No new draft email found with id starting with 'draft-'"

        # Check 3: selectedEmailId should be set to the new draft
        selected_email_id = final_state.get("selectedEmailId")
        if selected_email_id != new_draft.get("id"):
            return (
                0.0,
                f"Expected selectedEmailId to be '{new_draft.get('id')}', got '{selected_email_id}'",
            )

        return 1.0, "Successfully composed a new draft"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_fill_draft_fields(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully filled draft fields.

    This function checks:
    1. The draft email has to="teammate@example.com"
    2. The draft email has subject="Weekly update"
    3. The draft email has body containing "This is an update from me!"

    Args:
        initial_state: The initial state before filling
        final_state: The final state after filling

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Find the draft email
        emails = final_state.get("emails", [])
        draft_email = next((e for e in emails if e.get("id", "").startswith("draft-") and e.get("folder") == "drafts"), None)

        if draft_email is None:
            return 0.0, "No draft email found"

        # Check 1: to field should be "teammate@example.com"
        to_field = draft_email.get("to", "")
        if to_field != "teammate@example.com":
            return (
                0.0,
                f"Expected 'to' field to be 'teammate@example.com', got '{to_field}'",
            )

        # Check 2: subject should be "Weekly update"
        subject = draft_email.get("subject", "")
        if subject != "Weekly update":
            return (
                0.0,
                f"Expected subject to be 'Weekly update', got '{subject}'",
            )

        # Check 3: body should contain "This is an update from me!"
        body = draft_email.get("body", "")
        if "This is an update from me!" not in body:
            return (
                0.0,
                f"Expected body to contain 'This is an update from me!', got '{body}'",
            )

        return 1.0, "Successfully filled draft fields"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_send_the_draft(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully sent the draft.

    This function checks:
    1. The currentPage is "sent"
    2. A new sent email exists with id starting with "sent-"
    3. The sent email has to="teammate@example.com"
    4. The sent email has subject="Weekly update"
    5. The sent email has body="This is an update from me!"
    6. The sent email has folder="sent"

    Args:
        initial_state: The initial state before sending
        final_state: The final state after sending

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check 1: currentPage should be "sent"
        current_page = final_state.get("currentPage")
        if current_page != "sent":
            return (
                0.0,
                f"Expected currentPage to be 'sent', got '{current_page}'",
            )

        # Check 2: Find a new sent email
        emails = final_state.get("emails", [])
        sent_email = next((e for e in emails if e.get("id", "").startswith("sent-") and e.get("folder") == "sent"), None)

        if sent_email is None:
            return 0.0, "No new sent email found with id starting with 'sent-'"

        # Check 3: to field should be "teammate@example.com"
        to_field = sent_email.get("to", "")
        if to_field != "teammate@example.com":
            return (
                0.0,
                f"Expected 'to' field to be 'teammate@example.com', got '{to_field}'",
            )

        # Check 4: subject should be "Weekly update"
        subject = sent_email.get("subject", "")
        if subject != "Weekly update":
            return (
                0.0,
                f"Expected subject to be 'Weekly update', got '{subject}'",
            )

        # Check 5: body should be "This is an update from me!"
        body = sent_email.get("body", "")
        if body.strip() != "This is an update from me!":
            return (
                0.0,
                f"Expected body to be 'This is an update from me!', got '{body}'",
            )

        return 1.0, "Successfully sent the draft"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_discard_a_draft(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully discarded a draft.

    This function checks:
    1. Email with id "13" (Weekend plans draft) is removed from emails array
    2. The currentPage is "drafts"
    3. The selectedEmailId is null

    Args:
        initial_state: The initial state before discarding
        final_state: The final state after discarding

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check 1: Email with id "13" should not exist
        emails = final_state.get("emails", [])
        email_13 = next((e for e in emails if e.get("id") == "13"), None)

        if email_13 is not None:
            return 0.0, "Email with id '13' still exists, should be discarded"

        # Check 2: currentPage should be "drafts"
        current_page = final_state.get("currentPage")
        if current_page != "drafts":
            return (
                0.0,
                f"Expected currentPage to be 'drafts', got '{current_page}'",
            )

        # Check 3: selectedEmailId should be null
        selected_email_id = final_state.get("selectedEmailId")
        if selected_email_id is not None:
            return (
                0.0,
                f"Expected selectedEmailId to be null, got '{selected_email_id}'",
            )

        return 1.0, "Successfully discarded the draft"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_type_to_show_search_suggestions(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully typed to show search suggestions.

    This function checks:
    1. The searchQuery is "GitHub"
    2. The searchInputValue is "GitHub"
    3. The isSearchExpanded is true

    Args:
        initial_state: The initial state before typing
        final_state: The final state after typing

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check 1: searchQuery should be "GitHub"
        search_query = final_state.get("searchQuery", "")
        if search_query != "GitHub":
            return (
                0.0,
                f"Expected searchQuery to be 'GitHub', got '{search_query}'",
            )

        # Check 2: searchInputValue should be "GitHub"
        search_input_value = final_state.get("searchInputValue", "")
        if search_input_value != "GitHub":
            return (
                0.0,
                f"Expected searchInputValue to be 'GitHub', got '{search_input_value}'",
            )

        # Check 3: isSearchExpanded should be true
        is_search_expanded = final_state.get("isSearchExpanded")
        if not is_search_expanded:
            return (
                0.0,
                f"Expected isSearchExpanded to be true, got '{is_search_expanded}'",
            )

        return 1.0, "Successfully typed to show search suggestions"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_submit_a_search_with_enter(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully submitted a search with enter.

    This function checks:
    1. The currentPage is "search"
    2. The searchQuery is "user"
    3. The searchInputValue is "user"

    Args:
        initial_state: The initial state before submitting
        final_state: The final state after submitting

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check 1: currentPage should be "search"
        current_page = final_state.get("currentPage")
        if current_page != "search":
            return (
                0.0,
                f"Expected currentPage to be 'search', got '{current_page}'",
            )

        # Check 2: searchQuery should be "user"
        search_query = final_state.get("searchQuery", "")
        if search_query != "user":
            return (
                0.0,
                f"Expected searchQuery to be 'user', got '{search_query}'",
            )

        # Check 3: searchInputValue should be "user"
        search_input_value = final_state.get("searchInputValue", "")
        if search_input_value != "user":
            return (
                0.0,
                f"Expected searchInputValue to be 'user', got '{search_input_value}'",
            )

        return 1.0, "Successfully submitted search with enter"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_open_an_email_from_the_search_dropdown(
    initial_state: Dict[str, Any], final_state: Dict[str, Any]
) -> Tuple[float, str]:
    """
    Validate that the user successfully opened an email from the search dropdown.

    This function checks:
    1. The currentPage is "search"
    2. The selectedEmailId is "1"
    3. Email with id "1" has read status set to true

    Args:
        initial_state: The initial state before opening
        final_state: The final state after opening

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check 1: currentPage should be "search"
        current_page = final_state.get("currentPage")
        if current_page != "search":
            return (
                0.0,
                f"Expected currentPage to be 'search', got '{current_page}'",
            )

        # Check 2: selectedEmailId should be "1"
        selected_email_id = final_state.get("selectedEmailId")
        if selected_email_id != "1":
            return (
                0.0,
                f"Expected selectedEmailId to be '1', got '{selected_email_id}'",
            )

        # Check 3: Email with id "1" should have read = true
        emails = final_state.get("emails", [])
        email_1 = next((e for e in emails if e.get("id") == "1"), None)

        if email_1 is None:
            return 0.0, "Email with id '1' not found in emails array"

        if not email_1.get("read"):
            return (
                0.0,
                f"Expected email '1' to have read=true, got read={email_1.get('read')}",
            )

        return 1.0, "Successfully opened email from search dropdown"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_apply_the_has_attachments_filter(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully applied the has attachments filter.

    This function checks:
    1. The currentPage is "search"
    2. The searchFilters includes "hasAttachments"

    Args:
        initial_state: The initial state before applying filter
        final_state: The final state after applying filter

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check 1: currentPage should be "search"
        current_page = final_state.get("currentPage")
        if current_page != "search":
            return (
                0.0,
                f"Expected currentPage to be 'search', got '{current_page}'",
            )

        # Check 2: searchFilters should include "hasAttachments"
        search_filters = final_state.get("searchFilters", [])
        if "hasAttachments" not in search_filters:
            return (
                0.0,
                f"Expected searchFilters to include 'hasAttachments', got {search_filters}",
            )

        return 1.0, "Successfully applied has attachments filter"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_apply_the_unread_search_filter(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully applied the unread search filter.

    This function checks:
    1. The currentPage is "search"
    2. The searchFilters includes "unreads"

    Args:
        initial_state: The initial state before applying filter
        final_state: The final state after applying filter

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check 1: currentPage should be "search"
        current_page = final_state.get("currentPage")
        if current_page != "search":
            return (
                0.0,
                f"Expected currentPage to be 'search', got '{current_page}'",
            )

        # Check 2: searchFilters should include "unreads"
        search_filters = final_state.get("searchFilters", [])
        if "unreads" not in search_filters:
            return (
                0.0,
                f"Expected searchFilters to include 'unreads', got {search_filters}",
            )

        return 1.0, "Successfully applied unread search filter"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_choose_sent_items_from_the_folder_selector(
    initial_state: Dict[str, Any], final_state: Dict[str, Any]
) -> Tuple[float, str]:
    """
    Validate that the user successfully chose sent items from the folder selector.

    This function checks:
    1. The isSearchExpanded is true
    2. The selectedFolder is "sent"

    Args:
        initial_state: The initial state before selecting
        final_state: The final state after selecting

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check 1: isSearchExpanded should be true
        is_search_expanded = final_state.get("isSearchExpanded")
        if not is_search_expanded:
            return (
                0.0,
                f"Expected isSearchExpanded to be true, got '{is_search_expanded}'",
            )

        # Check 2: selectedFolder should be "sent"
        selected_folder = final_state.get("selectedFolder")
        if selected_folder != "sent":
            return (
                0.0,
                f"Expected selectedFolder to be 'sent', got '{selected_folder}'",
            )

        return 1.0, "Successfully chose sent items from folder selector"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_collapse_the_favorites_section(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully collapsed the favorites section.

    This function checks:
    1. The isFavoritesOpen is false

    Args:
        initial_state: The initial state before collapsing
        final_state: The final state after collapsing

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check: isFavoritesOpen should be false
        is_favorites_open = final_state.get("isFavoritesOpen")
        if is_favorites_open:
            return (
                0.0,
                f"Expected isFavoritesOpen to be false, got '{is_favorites_open}'",
            )

        return 1.0, "Successfully collapsed the favorites section"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_expand_the_navigation_sidebar(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully expanded the navigation sidebar.

    This function checks:
    1. The sidebarCollapsed is false

    Args:
        initial_state: The initial state before expanding
        final_state: The final state after expanding

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check: sidebarCollapsed should be false
        sidebar_collapsed = final_state.get("sidebarCollapsed")
        if sidebar_collapsed:
            return (
                0.0,
                f"Expected sidebarCollapsed to be false, got '{sidebar_collapsed}'",
            )

        return 1.0, "Successfully expanded the navigation sidebar"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


def _validate_collapse_the_navigation_sidebar(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """
    Validate that the user successfully collapsed the navigation sidebar.

    This function checks:
    1. The sidebarCollapsed is true

    Args:
        initial_state: The initial state before collapsing
        final_state: The final state after collapsing

    Returns:
        Tuple of (score, reason) where score is 1.0 for success, 0.0 for failure
    """
    try:
        logger.debug(f"Running reward function on state: {final_state}")

        # Check: sidebarCollapsed should be true
        sidebar_collapsed = final_state.get("sidebarCollapsed")
        if not sidebar_collapsed:
            return (
                0.0,
                f"Expected sidebarCollapsed to be true, got '{sidebar_collapsed}'",
            )

        return 1.0, "Successfully collapsed the navigation sidebar"

    except Exception as e:
        return 0.0, f"Error during validation: {str(e)}"


# Registry of all Outlook reward functions
REWARD_FUNCTIONS_OUTLOOK = {
    "_validate_go_to_the_drafts_page": _validate_go_to_the_drafts_page,
    "_validate_go_to_the_sent_items_page": _validate_go_to_the_sent_items_page,
    "_validate_open_an_email_from_the_list": _validate_open_an_email_from_the_list,
    "_validate_delete_an_email_from_the_list": _validate_delete_an_email_from_the_list,
    "_validate_flag_an_email": _validate_flag_an_email,
    "_validate_pin_an_email": _validate_pin_an_email,
    "_validate_toggle_read_on_an_email": _validate_toggle_read_on_an_email,
    "_validate_toggle_unread_on_an_email": _validate_toggle_unread_on_an_email,
    "_validate_compose_a_new_draft": _validate_compose_a_new_draft,
    "_validate_fill_draft_fields": _validate_fill_draft_fields,
    "_validate_send_the_draft": _validate_send_the_draft,
    "_validate_discard_a_draft": _validate_discard_a_draft,
    "_validate_type_to_show_search_suggestions": _validate_type_to_show_search_suggestions,
    "_validate_submit_a_search_with_enter": _validate_submit_a_search_with_enter,
    "_validate_open_an_email_from_the_search_dropdown": _validate_open_an_email_from_the_search_dropdown,
    "_validate_apply_the_has_attachments_filter": _validate_apply_the_has_attachments_filter,
    "_validate_apply_the_unread_search_filter": _validate_apply_the_unread_search_filter,
    "_validate_choose_sent_items_from_the_folder_selector": _validate_choose_sent_items_from_the_folder_selector,
    "_validate_collapse_the_favorites_section": _validate_collapse_the_favorites_section,
    "_validate_expand_the_navigation_sidebar": _validate_expand_the_navigation_sidebar,
    "_validate_collapse_the_navigation_sidebar": _validate_collapse_the_navigation_sidebar,
}



"""
Reward functions for Taobao app tasks.
"""

import logging
from typing import Any, Dict, Tuple

logger = logging.getLogger(__name__)

# Easy Tasks


def _validate_navigatetosearch(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that user navigated to search page."""

    # Check if we're on search page
    if final_state.get("page") != "search":
        return 0.0, f"Not on search page, current page: {final_state.get('page')}"

    # Search query should be empty (just navigated, not searched yet)
    if final_state.get("searchQuery") != "":
        return 0.0, f"Search query should be empty, got: {final_state.get('searchQuery')}"

    return 1.0, "Successfully navigated to search page"


def _validate_viewproductfromhome(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that user viewed product from homepage."""

    # Check if we're on preview page
    if final_state.get("page") != "preview":
        return 0.0, f"Not on preview page, current page: {final_state.get('page')}"

    # Check if a product is selected
    if not final_state.get("selectedProductId"):
        return 0.0, "No product selected"

    # Should have a valid product ID (numeric string)
    product_id = final_state.get("selectedProductId")
    if not product_id or not product_id.isdigit():
        return 0.0, f"Invalid product ID: {product_id}"

    return 1.0, f"Successfully viewing product {product_id} on preview page"


def _validate_viewshoppingcart(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that user navigated to cart page."""

    # Check if we're on cart page
    if final_state.get("page") != "cart":
        return 0.0, f"Not on cart page, current page: {final_state.get('page')}"

    # Cart should exist and have items (since prompt says "you have items")
    cart = final_state.get("cart", [])
    if len(cart) == 0:
        return 0.0, "Cart should contain items but is empty"

    return 1.0, "Successfully navigated to cart page with items"


def _validate_navigateviarelatedproducts(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that user navigated via related products."""

    # Check we're on product detail page
    if final_state.get("page") != "product":
        return 0.0, f"Should be on product detail page, got: {final_state.get('page')}"

    # Check correct product is selected (instant noodles - ID 3)
    if final_state.get("selectedProductId") != "3":
        return 0.0, f"Should be viewing instant noodles (product 3), got: {final_state.get('selectedProductId')}"

    return 1.0, "Successfully navigated to instant noodles via related products"


def _validate_browserelatedfromcart(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate browsing related products from cart page."""

    # Check we're on product page
    if final_state.get("page") != "product":
        return 0.0, f"Should be on preview page, got: {final_state.get('page')}"

    # Check a product is selected (any product from related)
    if not final_state.get("selectedProductId"):
        return 0.0, "No product selected from related products"

    # Should have a valid product ID
    product_id = final_state.get("selectedProductId")
    if not product_id or not product_id.isdigit():
        return 0.0, f"Invalid product ID selected: {product_id}"

    return 1.0, f"Successfully browsed to product {product_id} from cart related products"


def _validate_quickproductcomparison(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate quick product comparison using related products."""

    # Check we're on preview page for product 3
    if final_state.get("page") != "product":
        return 0.0, f"Should be on product page, got: {final_state.get('page')}"

    if final_state.get("selectedProductId") != "2":
        return 0.0, f"Should be viewing Jeep shirt (product 2), got: {final_state.get('selectedProductId')}"

    return 1.0, "Successfully compared multiple products using related products"


def _validate_returntosearchfromproduct(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate returning to search from product detail."""

    # Check we're on search page
    if final_state.get("page") != "search":
        return 0.0, f"Should be on search page, got: {final_state.get('page')}"

    # Search query should be preserved (not empty)
    search_query = final_state.get("searchQuery", "")
    if not search_query:
        return 0.0, "Search query should be preserved but is empty"

    return 1.0, f"Successfully returned to search with query: {search_query}"


def _validate_searchproduct(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that user searched for sauce products."""

    # Check we're on search page
    if final_state.get("page") != "search":
        return 0.0, f"Should be on search page, got: {final_state.get('page')}"

    # Check search query is "sauce"
    search_query = final_state.get("searchQuery", "").lower()
    if "sauce" not in search_query:
        return 0.0, f"Search query should contain 'sauce', got: {final_state.get('searchQuery')}"

    return 1.0, "Successfully searched for sauce products"


# Medium Tasks


def _validate_searchfindiphone(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that user searched for iPhone and found specific product."""

    # Check if we're on product page
    if final_state.get("page") != "product":
        return 0.0, f"Not on product page, current page: {final_state.get('page')}"

    # Check if correct product is selected
    if final_state.get("selectedProductId") != "1":
        return 0.0, f"Wrong product selected. Expected: 1, Got: {final_state.get('selectedProductId')}"

    # Check if search was performed with correct query
    search_query = final_state.get("searchQuery", "").lower()
    if "iphone" not in search_query:
        return 0.0, f"Search query doesn't contain 'iPhone': {final_state.get('searchQuery')}"

    return 1.0, "Successfully searched for iPhone and selected product 1"


def _validate_buynowfromdetails(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that product was added to cart from detail page."""

    # Check cart contains items
    cart = final_state.get("cart", [])
    if len(cart) == 0:
        return 0.0, "Cart is empty"

    # Find product with ID "2" in cart
    target_product = None
    for item in cart:
        if item.get("productId") == "2":
            target_product = item
            break

    if not target_product:
        return 0.0, f"Product 2 not found in cart. Cart contents: {cart}"

    # Check quantity is 1
    if target_product.get("qty", 0) != 1:
        return 0.0, f"Product 2 quantity should be 1, got: {target_product.get('qty')}"

    return 1.0, "Successfully added product 2 to cart with quantity 1"


def _validate_modifycartquantities(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that user modified cart quantities correctly."""

    # Check we're on cart page
    if final_state.get("page") != "cart":
        return 0.0, f"Should be on cart page, got: {final_state.get('page')}"

    cart = final_state.get("cart", [])

    product_3 = next((item for item in cart if item.get("productId") == "3"), None)
    product_7 = next((item for item in cart if item.get("productId") == "7"), None)

    if not product_3:
        return 0.0, "Instant noodles (product 3) not found in cart"
    if not product_7:
        return 0.0, "Laundry detergent (product 7) not found in cart"

    # Check updated quantities
    if product_3.get("qty", 0) != 3:
        return 0.0, f"Product 3 quantity should be 3, got: {product_3.get('qty')}"
    if product_7.get("qty", 0) != 1:
        return 0.0, f"Product 7 quantity should be 1, got: {product_7.get('qty')}"

    return 1.0, "Successfully modified cart quantities"


def _validate_useaddtocartpopup(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that user used add to cart popup successfully."""

    # Check add to cart popup is closed
    if final_state.get("isAddToCartPopupOpen", False):
        return 0.0, "Add to cart popup should be closed but is still open"

    # Check cart contains product 2 with quantity 2
    cart = final_state.get("cart", [])
    product_2 = next((item for item in cart if item.get("productId") == "2"), None)

    if not product_2:
        return 0.0, "JEEP shirt (product 2) not found in cart"

    if product_2.get("qty", 0) != 2:
        return 0.0, f"Product 2 quantity should be 2, got: {product_2.get('qty')}"

    return 1.0, "Successfully used add to cart popup to add 2 quantities"


def _validate_relatedproductchain(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate navigation through related product chain."""

    # Check we're on product detail page for sauce
    if final_state.get("page") != "product":
        return 0.0, f"Should be on product detail page, got: {final_state.get('page')}"

    if final_state.get("selectedProductId") != "5":
        return 0.0, f"Should be viewing sauce (product 5), got: {final_state.get('selectedProductId')}"

    return 1.0, "Successfully navigated through related product chain to sauce details"


def _validate_addrelatedtocart(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate adding related product to cart."""

    # Check cart contains product 4 with quantity 2
    cart = final_state.get("cart", [])
    product_4 = next((item for item in cart if item.get("productId") == "4"), None)

    if not product_4:
        return 0.0, "Cooking pan (product 4) not found in cart"

    if product_4.get("qty", 0) != 2:
        return 0.0, f"Product 4 quantity should be 2, got: {product_4.get('qty')}"

    # Add to cart popup should be closed
    if final_state.get("isAddToCartPopupOpen", False):
        return 0.0, "Add to cart popup should be closed"

    return 1.0, "Successfully added related product to cart with correct quantity"


# Hard Tasks


def _validate_complexsearchaddmultiple(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate complex search and multiple product actions."""

    # Check 1: Currently viewing product 4
    if final_state.get("page") != "product":
        return 0.0, f"Should be on product page, got: {final_state.get('page')}"

    if final_state.get("selectedProductId") != "4":
        return 0.0, f"Should be viewing product 4, got: {final_state.get('selectedProductId')}"

    # Check 2: Search query is "no"
    search_query = final_state.get("searchQuery", "").lower()
    if "no" not in search_query:
        return 0.0, f"Search query should contain 'no', got: {final_state.get('searchQuery')}"

    # Check 3: Product 3 is in cart
    cart = final_state.get("cart", [])
    product_3_in_cart = any(item.get("productId") == "3" for item in cart)

    if not product_3_in_cart:
        return 0.0, f"Product 3 not found in cart. Cart contents: {cart}"

    return 1.0, "Successfully completed complex search and multiple product actions"


def _validate_multiproductcartmanagement(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate complex multi-product cart management."""

    # Check 1: Currently on cart page
    if final_state.get("page") != "cart":
        return 0.0, f"Should be on cart page, got: {final_state.get('page')}"

    # Check 2: Cart contains both products with correct quantities
    cart = final_state.get("cart", [])

    product_7 = next((item for item in cart if item.get("productId") == "7"), None)
    product_3 = next((item for item in cart if item.get("productId") == "3"), None)

    if not product_7:
        return 0.0, "活力28洗衣液 (product 7) not found in cart"
    if not product_3:
        return 0.0, "华丰方便面 (product 3) not found in cart"

    # Check quantities
    if product_7.get("qty", 0) != 2:
        return 0.0, f"Product 7 quantity should be 2, got: {product_7.get('qty')}"
    if product_3.get("qty", 0) != 1:
        return 0.0, f"Product 3 quantity should be 1, got: {product_3.get('qty')}"

    return 1.0, "Successfully managed multiple products in cart with correct quantities"


def _validate_completeshoppingworkflow(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate complete shopping workflow with cart modification."""

    # Check 1: Currently on cart page
    if final_state.get("page") != "cart":
        return 0.0, f"Should be on cart page, got: {final_state.get('page')}"

    # Check 2: Cart contains only product 6 (massage chair)
    cart = final_state.get("cart", [])

    product_6 = next((item for item in cart if item.get("productId") == "6"), None)
    product_4 = next((item for item in cart if item.get("productId") == "4"), None)

    # Product 6 should be in cart with quantity 1
    if not product_6:
        return 0.0, "奥克斯按摩椅 (product 6) not found in cart"
    if product_6.get("qty", 0) != 1:
        return 0.0, f"Product 6 quantity should be 1, got: {product_6.get('qty')}"

    # Product 4 should NOT be in cart (was removed)
    if product_4:
        return 0.0, "爱仕达炒锅 (product 4) should be removed from cart but is still present"

    # Should have exactly 1 item in cart
    if len(cart) != 1:
        return 0.0, f"Cart should have exactly 1 item, got: {len(cart)} items"

    return 1.0, "Successfully completed shopping workflow with correct cart contents"


def _validate_completeproductdiscovery(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate complete product discovery flow."""

    # Check we're on cart page
    if final_state.get("page") != "cart":
        return 0.0, f"Should be on cart page, got: {final_state.get('page')}"

    cart = final_state.get("cart", [])

    product_5 = next((item for item in cart if item.get("productId") == "5"), None)
    product_7 = next((item for item in cart if item.get("productId") == "7"), None)

    if not product_5:
        return 0.0, "Sauce (product 5) not found in cart"
    if not product_7:
        return 0.0, "Laundry detergent (product 7) not found in cart"

    # Check quantities
    if product_5.get("qty", 0) != 2:
        return 0.0, f"Product 5 quantity should be 2, got: {product_5.get('qty')}"
    if product_7.get("qty", 0) != 1:
        return 0.0, f"Product 7 quantity should be 1, got: {product_7.get('qty')}"

    # Should have exactly 2 items in cart
    if len(cart) != 2:
        return 0.0, f"Cart should have exactly 2 items, got: {len(cart)} items"

    return 1.0, "Successfully completed product discovery flow"


def _validate_relatedproductsshoppingspree(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate complete shopping spree using related products."""

    # Check we're on cart page
    if final_state.get("page") != "cart":
        return 0.0, f"Should be on cart page, got: {final_state.get('page')}"

    cart = final_state.get("cart", [])

    # Check all products with correct quantities
    products_expected = {
        "4": 1,  # cooking pan
        "5": 2,  # sauce
        "3": 3,  # noodles
        "2": 1,  # detergent
    }

    for product_id, expected_qty in products_expected.items():
        product = next((item for item in cart if item.get("productId") == product_id), None)
        if not product:
            return 0.0, f"Product {product_id} not found in cart"
        if product.get("qty", 0) != expected_qty:
            return 0.0, f"Product {product_id} quantity should be {expected_qty}, got: {product.get('qty')}"

    # Should have exactly 4 items in cart
    if len(cart) != 4:
        return 0.0, f"Cart should have exactly 4 items, got: {len(cart)} items"

    return 1.0, "Successfully completed related products shopping spree with all items"


# Registry
REWARD_FUNCTIONS_TAOBAO_MOBILE = {
    "_validate_navigatetosearch": _validate_navigatetosearch,
    "_validate_viewproductfromhome": _validate_viewproductfromhome,
    "_validate_viewshoppingcart": _validate_viewshoppingcart,
    "_validate_navigateviarelatedproducts": _validate_navigateviarelatedproducts,
    "_validate_browserelatedfromcart": _validate_browserelatedfromcart,
    "_validate_quickproductcomparison": _validate_quickproductcomparison,
    "_validate_returntosearchfromproduct": _validate_returntosearchfromproduct,
    "_validate_searchproduct": _validate_searchproduct,
    "_validate_searchfindiphone": _validate_searchfindiphone,
    "_validate_buynowfromdetails": _validate_buynowfromdetails,
    "_validate_modifycartquantities": _validate_modifycartquantities,
    "_validate_useaddtocartpopup": _validate_useaddtocartpopup,
    "_validate_relatedproductchain": _validate_relatedproductchain,
    "_validate_addrelatedtocart": _validate_addrelatedtocart,
    "_validate_complexsearchaddmultiple": _validate_complexsearchaddmultiple,
    "_validate_multiproductcartmanagement": _validate_multiproductcartmanagement,
    "_validate_completeshoppingworkflow": _validate_completeshoppingworkflow,
    "_validate_completeproductdiscovery": _validate_completeproductdiscovery,
    "_validate_relatedproductsshoppingspree": _validate_relatedproductsshoppingspree,
}



"""Task definition and loading system."""

import string
import time
from abc import ABC, abstractmethod
from ast import Tuple
from typing import List, Optional

from datasets import Dataset, Features, Sequence, Value, load_dataset

from .dojos import Dojo
from .dojos.dojos_loader import load_dojos
from .dojos.rewards import get_reward_function
from .models import TaskDefinition, ValidateTask
from .task_models import RewardNotFoundError


class TaskLoader(ABC):
    """Abstract base class for task loaders."""

    @abstractmethod
    def load_task(self, task_path: str) -> TaskDefinition:
        """Load a task definition from a JSON file."""
        pass

    @abstractmethod
    def load_dojo_tasks(self, dojo_name: str) -> List[TaskDefinition]:
        """Load all tasks from a benchmark directory."""
        pass

    @abstractmethod
    def list_tasks(self, category: Optional[str] = None) -> dict[str, TaskDefinition]:
        """List all available tasks."""
        pass


class PyTaskLoader:
    """Loads tasks from python-based task registry"""

    @staticmethod
    def _register_all(bms: list[Dojo]) -> dict[string, Dojo]:
        result = {}
        for bm in bms:
            result[bm.get_id()] = bm

        return result

    def __init__(self):
        self.benchmarks_by_name = PyTaskLoader._register_all(load_dojos())

    def load_task(self, benchmark_task_path: str) -> TaskDefinition:
        split_path = benchmark_task_path.split("/")
        benchmark_name = split_path[0]
        task_id = split_path[1]
        tasks = self.load_benchmark_tasks(benchmark_name)
        for task in tasks:
            if task.id == task_id:
                return task
        raise ValueError(f"Task {task_id} not found in benchmark {benchmark_name}")

    def load_benchmark_tasks(self, benchmark_name: str) -> List[TaskDefinition]:
        return self.benchmarks_by_name[benchmark_name].get_tasks_list()

    def list_tasks(self, category: Optional[str] = None) -> dict[str, TaskDefinition]:
        raise NotImplementedError("List tasks not implemented")


class RemoteTaskLoader:
    """Loads tasks from a Hugging Face dataset."""

    def __init__(self, dataset_name: str, revision: str = "main"):
        self.dataset_name = dataset_name
        self.revision = revision
        self.dataset = None
        self.task_definitions = {}
        self.datasets_cache = {}
        self.tasks_without_reward = {}

    def _load_task_definitions(self):
        if self.task_definitions is None:
            self.task_definitions = self._get_all_tasks_definitions()
        return self.task_definitions

    def _load_dataset(self, name: str, revision: str) -> Dataset:
        """Lazy load the HF dataset with retry logic."""
        if name not in self.datasets_cache:
            features = Features(
                {
                    "spa": Value("string"),
                    "version": Value("string"),
                    "id": Value("string"),
                    "name": Value("string"),
                    "description": Value("string"),
                    "tier": Value("string"),
                    "initial_backend_state_name": Value("string"),
                    "instructions": Value("string"),
                    "reward_function": Value("string"),
                    "valid_target_states": Value("string"),
                    "max_steps": Value("int64"),
                    "timeout_seconds": Value("int64"),
                    "metadata": Value("string"),
                    "environment_type": Value("string"),
                    "trace_id": Value("string"),
                    "initial_state": Value("string"),
                    "environment": Value("string"),
                    "env_version": Value("string"),
                    "image": Value("string"),
                    "tools": Sequence(Value("string")),
                }
            )

            max_attempts = 5
            base_delay = 1  # seconds
            for attempt in range(max_attempts):
                try:
                    self.datasets_cache[name] = load_dataset(name, revision=revision, features=features)["train"]
                    print(f"✓ Loaded {len(self.datasets_cache[name])} tasks from HF dataset {name}")
                    break
                except Exception as e:
                    if attempt < max_attempts - 1:
                        delay = base_delay * (2**attempt)
                        print(f"Failed to load HF dataset (attempt {attempt + 1}/{max_attempts}): {e}")
                        print(f"Retrying in {delay} seconds...")
                        time.sleep(delay)
                    else:
                        print(f"Failed to load HF dataset after {max_attempts} attempts")
                        raise

        return self.datasets_cache[name]

    def _import_reward_function(self, function_name: str) -> ValidateTask:
        """Import a ValidateTask by name from the centralized rewards module."""
        if not function_name or function_name == "":
            raise RewardNotFoundError("Reward function name is empty")

        reward = get_reward_function(function_name)
        if reward is None:
            raise RewardNotFoundError(f"Reward function {function_name} not found")
        if not isinstance(reward, dict):
            raise RewardNotFoundError(f"Reward function {function_name} is not a ValidateTask")
        return reward

    def _get_all_task_definitions(self, dataset_name: str):
        """Load all tasks from HF dataset and convert to TaskDefinition objects."""
        if dataset_name not in self.datasets_cache:
            print(f"Loading tasks from dataset {self.dataset_name}")
            # TODO: Make this configurable
            dataset = self._load_dataset(dataset_name, revision=self.revision)

            self.task_definitions[dataset_name] = {}
            for row in dataset:
                # Use the clean constructor that handles HF-specific parsing
                try:
                    task = TaskDefinition.from_hf_row(row, reward_importer=self._import_reward_function)
                except RewardNotFoundError as e:
                    print(f"Skipping {row['spa']}/{row['id']} due to missing or invalid reward function {e}")
                    if dataset_name not in self.tasks_without_reward:
                        self.tasks_without_reward[dataset_name] = set()
                    self.tasks_without_reward[dataset_name].add(f"{row['spa']}/{row['id']}")
                    continue

                self.task_definitions[dataset_name][f"{task.spa}/{task.id}"] = task

            print(f"✓ Converted {len(self.task_definitions[dataset_name])} HF tasks to TaskDefinition objects")

        return self.task_definitions[dataset_name]

    def _get_all_tasks(self) -> List[TaskDefinition]:
        self.task_definitions = self._get_all_task_definitions("chakra-labs/dojo-bench-customer-colossus")
        return [task for task in self.task_definitions.values()]

    def load_benchmark_tasks(self, benchmark_name: str) -> Tuple(List[TaskDefinition], Optional[str]):
        dataset = self._load_dataset(benchmark_name, revision=self.revision)
        task_ids = [i["text"] for i in dataset]
        return ([self.load_task(task_id) for task_id in task_ids], dataset.version)

    def load_task(self, task_path: str) -> TaskDefinition:
        """
        Load a specific task by org/repo/dojo/task_id path or chakra-labs/dojo-bench-customer-colossus/dojo/task_id path
        if org/repo is not provided.
        """
        if "/" not in task_path:
            raise ValueError(f"Invalid task path format: {task_path}. Expected 'dojo/task_id'")

        parts = task_path.split("/")
        if len(parts) == 2:
            org = "chakra-labs"
            repo = "dojo-bench-customer-colossus"
            dojo_name = parts[0]
            task_id = parts[1]
        elif len(parts) == 4:
            org = parts[0]
            repo = parts[1]
            dojo_name = parts[2]
            task_id = parts[3]
        else:
            raise ValueError(f"Invalid task path format: {task_path}. Expected 'dojo/task_id' or 'org/repo/dojo/task_id'")

        if f"{org}/{repo}" not in self.task_definitions:
            self._get_all_task_definitions(f"{org}/{repo}")

        if f"{dojo_name}/{task_id}" in self.tasks_without_reward.get(f"{org}/{repo}", set()):
            raise RewardNotFoundError(
                f"reward for {task_path} not found." + " Please update your dojo-sdk-core package to the latest version",
            )

        if f"{dojo_name}/{task_id}" not in self.task_definitions[f"{org}/{repo}"]:
            raise ValueError(f"Task {task_path} not found in dataset {org}/{repo}")

        return self.task_definitions[f"{org}/{repo}"][f"{dojo_name}/{task_id}"]

    def list_tasks(self, category: Optional[str] = None) -> dict[str, TaskDefinition]:
        raise NotImplementedError("List tasks not implemented")



"""
Modular reward functions organized by game/application.

This package contains reward validation functions for different dojos,
organized into separate modules for maintainability.
"""

from dojo_sdk_core.dojos.rewards.backend import Backend
from dojo_sdk_core.dojos.rewards.canva_mcp import REWARD_FUNCTIONS_CANVA_MCP_V2
from dojo_sdk_core.dojos.rewards.figma_mcp import REWARD_FUNCTIONS_FIGMA_MCP_V2
from dojo_sdk_core.dojos.rewards.jd_v2 import REWARD_FUNCTIONS_JD_V2
from dojo_sdk_core.dojos.rewards.notion import REWARD_FUNCTIONS_NOTION_V2
from dojo_sdk_core.dojos.rewards.notion_database import REWARD_FUNCTIONS_NOTION_DATABASE_V2
from dojo_sdk_core.dojos.rewards.notion_gui import REWARD_FUNCTIONS_NOTION_GUI_V2
from dojo_sdk_core.dojos.rewards.notion_mixed import REWARD_FUNCTIONS_NOTION_MIXED_V2
from dojo_sdk_core.dojos.rewards.weibo_v2 import REWARD_FUNCTIONS_WEIBO_V2
from dojo_sdk_core.dojos.rewards.xiaohongshu_v2 import REWARD_FUNCTIONS_XIAOHONGSHU_V2

# Unified registry of all reward functions
REWARD_FUNCTIONS = {
    **REWARD_FUNCTIONS_JD_V2,
    **REWARD_FUNCTIONS_WEIBO_V2,
    **REWARD_FUNCTIONS_XIAOHONGSHU_V2,
    **REWARD_FUNCTIONS_NOTION_V2,
    **REWARD_FUNCTIONS_NOTION_DATABASE_V2,
    **REWARD_FUNCTIONS_NOTION_GUI_V2,
    **REWARD_FUNCTIONS_NOTION_MIXED_V2,
    **REWARD_FUNCTIONS_CANVA_MCP_V2,
    **REWARD_FUNCTIONS_FIGMA_MCP_V2,

}


def get_reward_function(name: str):
    """Get a reward function by name from the unified registry."""
    return REWARD_FUNCTIONS.get(name)


__all__ = ["REWARD_FUNCTIONS", "get_reward_function", "Backend"]



from enum import Enum
from typing import Annotated, List, Literal, Optional, Union

from pydantic import BaseModel, Discriminator

"""Type definitions for Dojo."""


class ActionType(str, Enum):
    KEY = "key"
    CLICK = "click"
    RIGHT_CLICK = "right_click"
    SCROLL = "scroll"
    TYPE = "type"
    DOUBLE_CLICK = "double_click"
    TRIPLE_CLICK = "triple_click"
    DRAG = "drag"
    MOVE_TO = "move_to"
    PRESS = "press"
    HOTKEY = "hotkey"
    MIDDLE_CLICK = "middle_click"
    DONE = "done"
    WAIT = "wait"


class KeyAction(BaseModel):
    type: Literal[ActionType.KEY] = ActionType.KEY
    key: str


class ClickAction(BaseModel):
    type: Literal[ActionType.CLICK] = ActionType.CLICK
    x: int
    y: int


class RightClickAction(BaseModel):
    type: Literal[ActionType.RIGHT_CLICK] = ActionType.RIGHT_CLICK
    x: int
    y: int


class ScrollAction(BaseModel):
    type: Literal[ActionType.SCROLL] = ActionType.SCROLL
    direction: str = "up"
    amount: int = 100


class TypeAction(BaseModel):
    type: Literal[ActionType.TYPE] = ActionType.TYPE
    text: str


class DoubleClickAction(BaseModel):
    type: Literal[ActionType.DOUBLE_CLICK] = ActionType.DOUBLE_CLICK
    x: int
    y: int


class TripleClickAction(BaseModel):
    type: Literal[ActionType.TRIPLE_CLICK] = ActionType.TRIPLE_CLICK
    x: int
    y: int


class DragAction(BaseModel):
    type: Literal[ActionType.DRAG] = ActionType.DRAG
    from_x: int
    from_y: int
    to_x: int
    to_y: int
    duration: float = 1.0


class MoveToAction(BaseModel):
    type: Literal[ActionType.MOVE_TO] = ActionType.MOVE_TO
    x: int
    y: int
    duration: float = 0.0


class PressAction(BaseModel):
    type: Literal[ActionType.PRESS] = ActionType.PRESS
    key: str


class HotkeyAction(BaseModel):
    type: Literal[ActionType.HOTKEY] = ActionType.HOTKEY
    keys: List[str]


class MiddleClickAction(BaseModel):
    type: Literal[ActionType.MIDDLE_CLICK] = ActionType.MIDDLE_CLICK
    x: int
    y: int


class DoneAction(BaseModel):
    type: Literal[ActionType.DONE] = ActionType.DONE
    response: Optional[str] = None


class WaitAction(BaseModel):
    type: Literal[ActionType.WAIT] = ActionType.WAIT
    seconds: int = 1


Action = Annotated[
    Union[
        KeyAction,
        ClickAction,
        RightClickAction,
        ScrollAction,
        TypeAction,
        DoubleClickAction,
        TripleClickAction,
        DragAction,
        MoveToAction,
        PressAction,
        HotkeyAction,
        MiddleClickAction,
        DoneAction,
        WaitAction,
    ],
    Discriminator("type"),
]


class Score(BaseModel):
    task_name: str
    score: float
    status: str
    success: bool
    steps_taken: int
    reward: float
    completion_reason: str



import json
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Tuple, TypedDict

from pydantic import BaseModel, Field

from dojo_sdk_core.dojos.rewards import REWARD_FUNCTIONS
from dojo_sdk_core.dojos.rewards.backend import Backend


class RewardNotFoundError(Exception):
    """Exception raised when a reward function is not found."""

    pass


# V2 reward types
class StateKeyQuery(TypedDict):
    collection: str
    filter: Dict[str, Any]


StateKey = Dict[str, StateKeyQuery]
ValidatorFunc = Callable[[Dict[str, Any], Dict[str, Any]], Tuple[float, str]]


class ValidateTask(TypedDict):
    state_key: StateKey
    validate_backend: ValidatorFunc
    validate_frontend: ValidatorFunc


class SettingsConfig(BaseModel):
    """Settings configuration."""

    anthropic_api_key: str = Field("", description="Anthropic API key")
    openai_api_key: str = Field("", description="OpenAI API key")
    openai_api_url: str = Field("", description="OpenAI API URL")
    dojo_websocket_endpoint: str = Field("", description="Dojo websocket endpoint")
    dojo_http_endpoint: str = Field("", description="Dojo http endpoint")
    posthog_api_key: str = Field("", description="PostHog API key")


class EnvironmentConfig(BaseModel):
    """Environment configuration."""

    type: str = Field(..., description="Environment type (e.g., 'spa')")
    path: str = Field(..., description="Path to environment file")


class InstructionsConfig(BaseModel):
    """Task instructions configuration."""

    user_prompt: str = Field(..., description="Prompt to show to the agent")
    success_criteria: str = Field(..., description="What constitutes success")


class TaskDefinition(BaseModel):
    spa: str = Field(..., description="SPA name")
    id: str = Field(..., description="Unique task identifier")
    name: str = Field(..., description="Human-readable task name")
    description: str = Field(..., description="Task description")
    version: str = Field(..., description="Task version")
    env_version: str = Field(..., description="Environment version")
    initial_backend_state_name: str = Field(..., description="Initial backend state/dataset name")
    environment: EnvironmentConfig = Field(..., description="Environment configuration")
    initial_state: dict[str, Any] = Field(..., description="Initial state for the environment")
    instructions: InstructionsConfig = Field(..., description="Task instructions")
    reward: Optional[ValidateTask] = Field(
        None,
        description="ValidateTask bundle with state_key, validate_backend, validate_frontend",
    )
    reward_v2: Optional[Callable[[Dict[str, Any], Backend], Tuple[float, str]]] = Field(
        None,
        description="Reward function for v2 tasks",
    )
    reward_function_name: Optional[str] = Field(
        None,
        description="Name of the reward function (e.g., 'notion/task_create_page') for verifier service",
    )
    max_steps: int = Field(default=10, description="Maximum number of steps allowed")
    timeout_seconds: int = Field(default=60, description="Task timeout in seconds")
    metadata: dict[str, Any] = Field(default_factory=dict, description="Additional metadata")
    environment_type: str = Field(default="gui", description="Environment type e.g 'mcp' or 'gui'")
    trace_id: Optional[str] = Field(None, description="Unique trace identifier for the task")

    def model_post_init(self, __context: Any) -> None:
        """Validate that reward is provided."""
        if self.version == "2.0" and not self.reward_v2:
            raise ValueError("reward_v2 must be provided for v2 tasks")
        if self.version == "1.0" and not self.reward:
            raise ValueError("reward must be provided for v1 tasks")

    def get_environment_path(self, base_path: Optional[Path] = None) -> Path:
        """Get absolute path to environment file."""
        if base_path is None:
            base_path = Path.cwd()
        return base_path / self.environment.path

    @classmethod
    def from_hf_row(
        cls, row: dict[str, Any], reward_importer: Optional[Callable[[str], Optional[ValidateTask]]] = None
    ) -> "TaskDefinition":
        """Create a TaskDefinition from a HuggingFace dataset row."""
        initial_state = json.loads(row["initial_state"])
        environment = json.loads(row["environment"])
        instructions = json.loads(row["instructions"])
        metadata = json.loads(row["metadata"])
        version = row["version"]
        reward_v2 = None
        reward = None
        environment_type = row.get("environment_type")
        if not environment_type:
            print(f"environment_type is empty for task {row['id']}, setting to default 'gui'")
            environment_type = "gui"

        reward_function_name = row.get("reward_function", "").strip() if row.get("reward_function") else None

        if version == "2.0":
            reward_v2 = REWARD_FUNCTIONS.get(row["reward_function"].strip())
            if not reward_v2:
                raise RewardNotFoundError(f"Reward function {row['reward_function']} not found")
        elif version == "1.0" and row["reward_function"] and row["reward_function"].strip():
            if reward_importer:
                reward = reward_importer(row["reward_function"])

        return cls(
            id=row["id"],
            spa=row["spa"],
            name=row["name"],
            description=row["description"],
            environment=environment,
            initial_state=initial_state,
            version=version,
            env_version=row.get("env_version", ""),
            initial_backend_state_name=row.get("initial_backend_state_name", ""),
            instructions=instructions,
            reward=reward,
            reward_v2=reward_v2,
            reward_function_name=reward_function_name,
            max_steps=row["max_steps"],
            timeout_seconds=row["timeout_seconds"],
            metadata=metadata,
            environment_type=environment_type,
            trace_id=row.get("trace_id"),
        )



"""
Reward functions for Gmail app tasks.
"""

import logging
from typing import Any, Dict, Tuple

logger = logging.getLogger(__name__)


def _validate_changing_categories_in_inbox(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that categories were changed in the inbox."""
    if "activeCategory" not in final_state:
        return 0.0, "No activeCategory in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    # Check that activeCategory changed from primary to updates
    if final_state.get("activeCategory") == "updates":
        return 1.0, "Successfully changed category to updates"

    return (
        0.0,
        "Expected activeCategory to be 'updates', got '{}'".format(final_state.get("activeCategory")),
    )


def _validate_collapse_the_sidebar(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that the sidebar was collapsed."""
    if "sidebarCollapsed" not in final_state:
        return 0.0, "No sidebarCollapsed in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    if final_state.get("sidebarCollapsed") is True:
        return 1.0, "Sidebar successfully collapsed"

    return (
        0.0,
        "Expected sidebarCollapsed to be True, got '{}'".format(final_state.get("sidebarCollapsed")),
    )


def _validate_expand_sidebar(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that the sidebar was expanded."""
    if "sidebarCollapsed" not in final_state:
        return 0.0, "No sidebarCollapsed in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    if final_state.get("sidebarCollapsed") is False:
        return 1.0, "Sidebar successfully expanded"

    return (
        0.0,
        "Expected sidebarCollapsed to be False, got '{}'".format(final_state.get("sidebarCollapsed")),
    )


def _validate_filter_using_search_and_sender(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that search filter was applied using search query and sender."""
    if "emails" not in final_state or "selectedEmailId" not in final_state:
        return 0.0, "No emails or selectedEmailId in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    selected_email_id = final_state.get("selectedEmailId")
    if not selected_email_id:
        return 0.0, "No email selected"

    # Find the selected email
    selected_email = next(
        (email for email in final_state["emails"] if email.get("id") == selected_email_id),
        None,
    )
    if not selected_email:
        return 0.0, f"Selected email {selected_email_id} not found in emails"

    # Check that the email is from GitHub and has the correct subject
    sender_name = selected_email.get("sender", {}).get("name", "")
    subject = selected_email.get("subject", "")

    if sender_name == "GitHub" and "Your recent order confirmation" in subject:
        return (
            1.0,
            "Successfully filtered and selected email from GitHub with subject 'Your recent order confirmation'",
        )

    errors = []
    if sender_name != "GitHub":
        errors.append(f"sender={sender_name} (expected 'GitHub')")
    if "Your recent order confirmation" not in subject:
        errors.append("subject does not contain 'Your recent order confirmation'")

    return 0.0, "; ".join(errors) if errors else "Email selection validation failed"


def _validate_filtering_apple_emails_with_attachment(
    initial_state: Dict[str, Any], final_state: Dict[str, Any]
) -> Tuple[float, str]:
    """Validate that Apple emails with attachments were filtered."""
    if "emails" not in final_state or "selectedEmailId" not in final_state:
        return 0.0, "No emails or selectedEmailId in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    selected_email_id = final_state.get("selectedEmailId")
    if not selected_email_id:
        return 0.0, "No email selected"

    # Find the selected email
    selected_email = next(
        (email for email in final_state["emails"] if email.get("id") == selected_email_id),
        None,
    )
    if not selected_email:
        return 0.0, f"Selected email {selected_email_id} not found in emails"

    # Check email properties
    sender_name = selected_email.get("sender", {}).get("name", "")
    subject = selected_email.get("subject", "")
    body = selected_email.get("body", "")
    expected_text = "I dropped a concise recap in the shared channel so the broader group can keep pace"

    errors = []
    if "Preview the latest product tour" not in subject:
        errors.append("subject does not contain 'Preview the latest product tour'")
    if sender_name != "Apple":
        errors.append(f"sender={sender_name} (expected 'Apple')")
    if expected_text not in body:
        errors.append(f"body does not contain expected text: '{expected_text}'")

    if errors:
        return 0.0, "; ".join(errors)

    return (
        1.0,
        "Successfully filtered and selected Apple email with attachment containing expected text",
    )


def _validate_filtering_your_own_sent_emails(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that own sent emails were filtered."""
    if "emails" not in final_state or "selectedEmailId" not in final_state:
        return 0.0, "No emails or selectedEmailId in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    selected_email_id = final_state.get("selectedEmailId")
    if not selected_email_id:
        return 0.0, "No email selected"

    # Find the selected email
    selected_email = next(
        (email for email in final_state["emails"] if email.get("id") == selected_email_id),
        None,
    )
    if not selected_email:
        return 0.0, f"Selected email {selected_email_id} not found in emails"

    # Check that the email is from Maximilian Falco and has the correct subject
    sender_name = selected_email.get("sender", {}).get("name", "")
    subject = selected_email.get("subject", "")

    if sender_name == "Maximilian Falco" and "Q4 roadmap outline" in subject:
        return (
            1.0,
            "Successfully filtered and selected sent email from Maximilian Falco with subject 'Q4 roadmap outline'",
        )

    errors = []
    if sender_name != "Maximilian Falco":
        errors.append(f"sender={sender_name} (expected 'Maximilian Falco')")
    if "Q4 roadmap outline" not in subject:
        errors.append("subject does not contain 'Q4 roadmap outline'")

    return 0.0, "; ".join(errors) if errors else "Email selection validation failed"


def _validate_focus_a_single_email(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that a single email was focused."""
    if "emails" not in final_state or "selectedEmailId" not in final_state:
        return 0.0, "No emails or selectedEmailId in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    selected_email_id = final_state.get("selectedEmailId")
    if not selected_email_id:
        return 0.0, "No email selected"

    # Find the selected email
    selected_email = next(
        (email for email in final_state["emails"] if email.get("id") == selected_email_id),
        None,
    )
    if not selected_email:
        return 0.0, f"Selected email {selected_email_id} not found in emails"

    # Check that an email is selected (the focus state reflects the email)
    if selected_email_id == "mail-3":
        return 1.0, f"Successfully focused email {selected_email_id}"

    return (
        0.0,
        f"Expected selectedEmailId to be 'mail-3', got '{selected_email_id}'",
    )


def _validate_navigate_home(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that navigation to home (inbox primary category) occurred."""
    if "activeCategory" not in final_state:
        return 0.0, "No activeCategory in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    if final_state.get("activeCategory") == "primary":
        return 1.0, "Successfully navigated to inbox primary category"

    return (
        0.0,
        f"Expected activeCategory to be 'primary', got '{final_state.get('activeCategory')}'",
    )


def _validate_navigate_to_different_categories_via_sidebar(
    initial_state: Dict[str, Any], final_state: Dict[str, Any]
) -> Tuple[float, str]:
    """Validate that navigation to Sent category via sidebar occurred."""
    if "emails" not in final_state:
        return 0.0, "No emails in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    # Check that we're viewing sent emails (emails where sent=True)
    sent_emails = [email for email in final_state["emails"] if email.get("sent") is True]

    # The task expects to reach Sent category and see different content
    # We verify by checking that sent emails are present and the list is different from inbox
    if len(sent_emails) > 0:
        return (
            1.0,
            f"Successfully navigated to Sent category via sidebar, showing {len(sent_emails)} sent emails",
        )

    return 0.0, "No sent emails found in final state"


def _validate_navigate_to_the_starred_category(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that navigation to starred category occurred."""
    if "shortcuts" not in final_state:
        return 0.0, "No shortcuts in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    # Check that the starred shortcut is active
    starred_shortcut = next((s for s in final_state.get("shortcuts", []) if s.get("id") == "starred"), None)

    if not starred_shortcut:
        return 0.0, "Starred shortcut not found in shortcuts"

    if starred_shortcut.get("isActive") is True:
        return 1.0, "Successfully navigated to starred category via sidebar"

    return (
        0.0,
        f"Starred shortcut is not active. isActive={starred_shortcut.get('isActive')}",
    )


def _validate_open_filter_modal(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that the filter modal was opened."""
    if "isSearchFilterOpen" not in final_state:
        return 0.0, "No isSearchFilterOpen in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    if final_state.get("isSearchFilterOpen") is True:
        return 1.0, "Filter modal successfully opened"

    return (
        0.0,
        f"Expected isSearchFilterOpen to be True, got '{final_state.get('isSearchFilterOpen')}'",
    )


def _validate_see_sent_emails(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that sent emails are being viewed."""
    if "emails" not in final_state:
        return 0.0, "No emails in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    # Check that all visible emails are sent by Maximilian Falco
    sent_emails = [email for email in final_state["emails"] if email.get("sent") is True]
    all_from_user = all(email.get("sender", {}).get("name") == "Maximilian Falco" for email in sent_emails)

    if len(sent_emails) > 0 and all_from_user:
        return (
            1.0,
            f"Successfully viewing sent emails, showing {len(sent_emails)} emails from Maximilian Falco",
        )

    errors = []
    if len(sent_emails) == 0:
        errors.append("No sent emails found")
    if not all_from_user:
        errors.append("Not all emails are from Maximilian Falco")

    return 0.0, "; ".join(errors) if errors else "Sent emails validation failed"


def _validate_sending_an_email(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that an email was sent."""
    if "emails" not in final_state:
        return 0.0, "No emails in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    # Count sent emails
    initial_sent_count = len([email for email in initial_state.get("emails", []) if email.get("sent")])
    final_sent_count = len([email for email in final_state["emails"] if email.get("sent")])

    # Check that sent count increased by 1
    if final_sent_count == initial_sent_count + 1:
        # Find the email with subject "Test"
        test_email = next(
            (email for email in final_state["emails"] if email.get("sent") and email.get("subject") == "Test"),
            None,
        )

        if test_email:
            body = test_email.get("body", "")
            if body == "Test":
                return (
                    1.0,
                    f"Successfully sent email. Sent count increased from {initial_sent_count} to {final_sent_count}. Email has subject 'Test' and body 'Test'",  # noqa: E501
                )
            else:
                return (
                    0.0,
                    f"Email with subject 'Test' found but body is '{body}' (expected 'Test')",
                )

        return 0.0, "Email with subject 'Test' not found in sent emails"

    return (
        0.0,
        f"Sent count did not increase by 1. Initial: {initial_sent_count}, Final: {final_sent_count}",
    )


def _validate_snooze_a_single_email(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that a single email was snoozed."""
    if "emails" not in final_state:
        return 0.0, "No emails in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    # Find snoozed emails
    snoozed_emails = [email for email in final_state["emails"] if email.get("snoozed") is True]

    # Check that exactly one email is snoozed, and it's from GitHub with the correct subject
    if len(snoozed_emails) == 1:
        snoozed_email = snoozed_emails[0]
        sender_name = snoozed_email.get("sender", {}).get("name", "")
        subject = snoozed_email.get("subject", "")

        if sender_name == "GitHub" and "Your recent order confirmation" in subject:
            return (
                1.0,
                "Successfully snoozed email from GitHub titled 'Your recent order confirmation'",
            )

        errors = []
        if sender_name != "GitHub":
            errors.append(f"sender={sender_name} (expected 'GitHub')")
        if "Your recent order confirmation" not in subject:
            errors.append("subject does not contain 'Your recent order confirmation'")
        return 0.0, "; ".join(errors)

    return (
        0.0,
        "Expected exactly 1 snoozed email, found {}".format(len(snoozed_emails)),
    )


def _validate_starring_a_non_primary_email(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that a non-primary email was starred."""
    if "emails" not in final_state:
        return 0.0, "No emails in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    # Find the GitHub email from updates category that should be starred
    github_email = next(
        (
            email
            for email in final_state["emails"]
            if email.get("sender", {}).get("name") == "GitHub" and email.get("category") == "updates"
        ),
        None,
    )

    if not github_email:
        return 0.0, "GitHub email from updates category not found"

    # Check that it's starred
    # When an email is starred, it appears in the primary category view even if its category is "updates"
    # This is handled by the UI filtering logic: starred emails show in primary category
    if github_email.get("starred") is True:
        return (
            1.0,
            "Successfully starred GitHub email from updates category. Email now appears in primary category view (starred emails are visible in primary)",  # noqa: E501
        )

    return (
        0.0,
        f"GitHub email from updates category is not starred. starred={github_email.get('starred')}",
    )


def _validate_starring_an_email(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that an email was starred."""
    if "emails" not in final_state:
        return 0.0, "No emails in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    # Find the Twitter email with subject "Product roadmap highlights"
    target_email = next(
        (
            email
            for email in final_state["emails"]
            if email.get("sender", {}).get("name") == "Twitter" and "Product roadmap highlights" in email.get("subject", "")
        ),
        None,
    )

    if not target_email:
        return (
            0.0,
            "Email from Twitter with subject 'Product roadmap highlights' not found",
        )

    # Check that it's starred
    if target_email.get("starred") is True:
        return (
            1.0,
            "Successfully starred email from Twitter titled 'Product roadmap highlights'",
        )

    return (
        0.0,
        f"Email from Twitter is not starred. starred={target_email.get('starred')}",
    )


def _validate_trigger_filter_bar(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that the filter bar was triggered."""
    # The filter bar appearing is typically indicated by searchFilters being active
    # or isSearchFilterOpen being true, but let's check what the actual state shows
    logger.debug(f"Running reward function on state: {final_state}")

    # Check if searchFilters exist and are being used
    if "searchFilters" in final_state:
        # The filter bar is triggered when search filters are active
        # This could be indicated by isSearchFilterOpen or by filters having values
        if final_state.get("isSearchFilterOpen") is True:
            return 1.0, "Filter bar successfully triggered and visible"

    # Alternative: check if search query or filters are set
    if final_state.get("searchQuery") or final_state.get("activeSearchQuery"):
        return 1.0, "Filter bar successfully triggered via search"

    return (
        0.0,
        "Filter bar not triggered. isSearchFilterOpen is not True and no search query is set",
    )


def _validate_unsnooze_all_emails(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that all emails were unsnoozed."""
    if "emails" not in final_state:
        return 0.0, "No emails in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    # Check that no emails are snoozed
    snoozed_emails = [email for email in final_state["emails"] if email.get("snoozed") is True]

    if len(snoozed_emails) == 0:
        return 1.0, "Successfully unsnoozed all emails. No snoozed emails remaining"

    return (
        0.0,
        f"Expected 0 snoozed emails, found {len(snoozed_emails)}",
    )


def _validate_unstar_all_emails(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that all emails were unstarred."""
    if "emails" not in final_state:
        return 0.0, "No emails in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    # Check that no emails are starred
    starred_emails = [email for email in final_state["emails"] if email.get("starred") is True]

    if len(starred_emails) == 0:
        return 1.0, "Successfully unstarred all emails. No starred emails remaining"

    return (
        0.0,
        f"Expected 0 starred emails, found {len(starred_emails)}",
    )


def _validate_unstarring_an_email(initial_state: Dict[str, Any], final_state: Dict[str, Any]) -> Tuple[float, str]:
    """Validate that an email was unstarred."""
    if "emails" not in final_state:
        return 0.0, "No emails in final state"

    logger.debug(f"Running reward function on state: {final_state}")

    # Find the GitHub email
    github_email = next(
        (email for email in final_state["emails"] if email.get("sender", {}).get("name") == "GitHub"),
        None,
    )

    if not github_email:
        return 0.0, "GitHub email not found"

    # Check that it's not starred
    if github_email.get("starred") is False:
        # Also verify it's not visible in primary category if it was moved
        return 1.0, "Successfully unstarred GitHub email. Email is no longer starred"

    return (
        0.0,
        f"GitHub email is still starred. starred={github_email.get('starred')}",
    )


# Registry of all Gmail reward functions
REWARD_FUNCTIONS_GMAIL = {
    "_validate_changing_categories_in_inbox": _validate_changing_categories_in_inbox,
    "_validate_collapse_the_sidebar": _validate_collapse_the_sidebar,
    "_validate_expand_sidebar": _validate_expand_sidebar,
    "_validate_filter_using_search_and_sender": _validate_filter_using_search_and_sender,
    "_validate_filtering_apple_emails_with_attachment": _validate_filtering_apple_emails_with_attachment,
    "_validate_filtering_your_own_sent_emails": _validate_filtering_your_own_sent_emails,
    "_validate_focus_a_single_email": _validate_focus_a_single_email,
    "_validate_navigate_home": _validate_navigate_home,
    "_validate_navigate_to_different_categories_via_sidebar": _validate_navigate_to_different_categories_via_sidebar,
    "_validate_navigate_to_the_starred_category": _validate_navigate_to_the_starred_category,
    "_validate_open_filter_modal": _validate_open_filter_modal,
    "_validate_see_sent_emails": _validate_see_sent_emails,
    "_validate_sending_an_email": _validate_sending_an_email,
    "_validate_snooze_a_single_email": _validate_snooze_a_single_email,
    "_validate_starring_a_non_primary_email": _validate_starring_a_non_primary_email,
    "_validate_starring_an_email": _validate_starring_an_email,
    "_validate_trigger_filter_bar": _validate_trigger_filter_bar,
    "_validate_unsnooze_all_emails": _validate_unsnooze_all_emails,
    "_validate_unstar_all_emails": _validate_unstar_all_emails,
    "_validate_unstarring_an_email": _validate_unstarring_an_email,
}



from dojo_sdk_core.task_models import (
    EnvironmentConfig,
    InstructionsConfig,
    SettingsConfig,
    StateKey,
    StateKeyQuery,
    TaskDefinition,
    ValidateTask,
    ValidatorFunc,
)

__all__ = [
    "EnvironmentConfig",
    "InstructionsConfig",
    "SettingsConfig",
    "StateKey",
    "StateKeyQuery",
    "TaskDefinition",
    "ValidateTask",
    "ValidatorFunc",
]
